/*
 * Copyright (c) 2001 by Pankaj Kumar.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed. The LICENSE or reference to the LICENSE
 * can be found in file LICENSE.txt in the base directory of distributed
 * software.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */
package org.xperf.xpb;

import java.io.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import org.xml.sax.helpers.DefaultHandler;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.Iterator;
import java.util.Date;

public class XPBDriver extends DefaultHandler
  {
  final static String PDATA_FILE = "pdata.xml";
  final static String RESULTS_FILE = "results.xml";

  private HashMap pActivityMap = null;
  private Vector argPMethods = null;
  private ArgsReader ardr = null;

  void usage()
    {
    System.err.print(
	  "Usage\n"
		+"  java org.xperf.xpb.XPBDriver [options] <target1> <target2> ...\n"
		+"\n"
		+"Options:\n"
		+"  -display    Display performance metrics. Default: on\n"
		+"  -verbose    Verbose output. Shows execution progress. Default: off\n"
		+"  -gc         Force garbage collection before each run. Default: off\n"
		+"  -gcMeasured Force garbage collection within measurement window. Default: off\n"
		+"  -loopCount COUNT number of iteration loops. Default: 5.\n"
		+"\n"
		+"This program performs specified processing on a set of target,\n"
		+"\n"
    );
    System.exit(1);
	}

  private void parseCLArgs(String[] args, ArgsReader ardr)
    {
    for (int i = 0; i < args.length; ++i)
      {
	  String arg = args[i];
	  String value = null;
	  if (i < args.length - 1)
	    value = args[i + 1];

	  if (arg.charAt(0) == '-')
	    {
		if (arg.equals("-help"))
		  usage();
		else if (arg.equals("-display"))
		  ardr.displayFlag = true;
		else if (arg.equals("-verbose"))
		  ardr.verboseFlag = true;
		else if (arg.equals("-gc"))
		  ardr.gcFlag = true;
		else if (arg.equals("-gcMeasured"))
		  ardr.gcMeasuredFlag = true;
		else if (arg.equals("-loopCount"))
		  {
		  try
		    {
		    ardr.loopCount = (Integer.decode(value)).intValue();
		    ++i;
		    }
		  catch (NumberFormatException e)
		    {
			System.err.println("Invalid Loop Count specified. Exception: " + e);
			System.err.println("Resorting to default value of: " + ardr.loopCount);
		    }
	      }
		else
		  usage();
	    }
	  else
		ardr.uris.add(arg);
      }

	if (ardr.uris.size() == 0)
	  {
	  System.err.println("Missing input file.");
	  System.exit(1);
	  }
    }

  private void display()
    {
	int loopCount = ardr.loopCount;
	int fsize = 8;
	int jstep = (ardr.loopCount/7) + 1;

	StringBuffer hdr = new StringBuffer();
	StringBuffer hr = new StringBuffer();

	addField(hdr, "method", fsize);
	hr.append("--------");
	for (int j = 0; j < loopCount; j += jstep)
	  {
	  addField(hdr, "Run#" + Integer.toString(j), fsize);
	  hr.append("--------");
	  }
	addField(hdr, "Average", fsize);
	hr.append("--------");

	System.out.println("Execution Time (in milli seconds)::");
	System.out.println(hdr.toString());
	System.out.println(hr.toString());
	for (int i = 0; i < argPMethods.size(); i++)
	  {
	  PMethod pm = (PMethod)argPMethods.elementAt(i);
	  StringBuffer sb = new StringBuffer();

	  addField(sb, pm.name, fsize);
	  for (int j = 0; j < loopCount; j += jstep)
		addField(sb, Long.toString(pm.execTime[j]), fsize);

	  long total = 0;
	  for (int j = 1; j < loopCount; j++)
		total += pm.execTime[j];
	  long average = total/(loopCount - 1);
      addField(sb, Long.toString(average), fsize);

	  System.out.println(sb.toString());
	  System.out.println(hr.toString());
	  }

	System.out.println("\nMemory Used (in KB)::");
	System.out.println(hdr.toString());
	System.out.println(hr.toString());
	for (int i = 0; i < argPMethods.size(); i++)
	  {
      PMethod pm = (PMethod)argPMethods.elementAt(i);
	  StringBuffer sb = new StringBuffer();

	  addField(sb, pm.name, fsize);
	  for (int j = 0; j < loopCount; j += jstep)
		addField(sb, Long.toString(pm.memUsed[j]/1024), fsize);

	  System.out.println(sb.toString());
	  System.out.println(hr.toString());
	  }
	}

  private void addField(StringBuffer sb, String fld, int fsize)
    {
	sb.append(fld);
	if (fld.length() < fsize)
	  {
	  for (int i = 0; i < fsize - fld.length(); i++)
	    sb.append(" ");
      }
	}

  private void configure()
    {
	// Read the conf.xml file for all processing activities and methods.
	ConfReader crdr = new ConfReader();
	pActivityMap = crdr.configure();
	}

  private void createProcessors()
    {
	for (int i = 0; i < argPMethods.size(); i++)
	  {
	  PMethod pm = (PMethod)argPMethods.elementAt(i);
	  try
	    {
	    Class classFor = Class.forName(pm.className);
	    pm.processor = (Processor)classFor.newInstance();
	    if (pm.args != null)
	      pm.processor.setup(pm.args);
	    }
	  catch (Exception e)
	    {
		System.err.println("Processor creation failed for (" + pm.paName + ", " + pm.name + ") : " + e);
		argPMethods.remove(i);	// Exclude it from processing !!
		--i;
	    }
	  }
	}

  private void process()
    {

	for (int i = 0; i < argPMethods.size(); i++)
	  {
	  PMethod pm = (PMethod)argPMethods.elementAt(i);
	  try
	    {
	    pm.process(ardr.loopCount, ardr.uris, ardr.gcFlag, ardr.gcMeasuredFlag);
	    }
	  catch (Exception e)
	    {
		System.err.println("processing failed for (" + pm.paName + ", " + pm.name + ") : " + e);
		e.printStackTrace(System.err);
		argPMethods.remove(i);	// Exclude it from reporting !!
		--i;
		}
      }
    }

  private void writePerfData(String fileName)
    {
	try
	  {
	  BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	  StringBuffer sb = new StringBuffer();

	  sb.append("<?xml version = \"1.0\"?>\n");
	  sb.append("<!-- XPB Execution Performance Data -->\n");
	  sb.append("<xpbpdata xmlns=\"http://www.xperf.org/xpb\">\n");
	  sb.append("  <SystemInfo>\n");

	  sb.append("    <JavaInfo java.version=\"" + System.getProperty("java.version") + "\" ");
	  sb.append("java.vm.version=\"" + System.getProperty("java.vm.version") + "\" ");
	  sb.append("java.vendor=\"" + System.getProperty("java.vendor") + "\"/>\n");

	  sb.append("    <OSInfo os.name=\"" + System.getProperty("os.name") + "\" ");
	  sb.append("os.arch=\"" + System.getProperty("os.arch") + "\" ");
	  sb.append("os.version=\"" + System.getProperty("os.version") + "\"/>\n");

	  sb.append("  </SystemInfo>\n");

	  sb.append("  <MeasurementTime millis=\"" + System.currentTimeMillis() + "\"" +
	            " timestamp=\"" + new Date() + "\"/>\n");

	  sb.append("  <Targets>\n");
	  for (int i = 0; i < ardr.uris.size(); i++)
	    {
		String name = (String)ardr.uris.elementAt(i);
		File f = new File(name);
	    sb.append("    <Target name=\"" + name + "\" sizeKB=\"" + (f.length()/1024) + "\"/>\n");
	    }
	  sb.append("  </Targets>\n");

	  sb.append("  <Flags gc=\"" + ardr.gcFlag + "\" gcMeasured=\"" + ardr.gcMeasuredFlag + "\"/>\n");
	  sb.append("  <Params loopCount=\"" + ardr.loopCount + "\"/>\n");

      sb.append("  <PerformanceData>\n");
	  bw.write(sb.toString(), 0, sb.length());
	  bw.flush();

	  for (int i = 0; i < argPMethods.size(); i++)
	    {
	    PMethod pm = (PMethod)argPMethods.elementAt(i);
	    String pData = pm.perfDataAsXML();
	    bw.write(pData, 0, pData.length());
	    bw.flush();
        }

      sb = new StringBuffer();
      sb.append("  </PerformanceData>\n");
      sb.append("</xpbpdata>\n");
      bw.write(sb.toString(), 0, sb.length());
      bw.close();
      }
    catch (IOException ioe)
      {
	  System.err.println("Encountered error in writing performance data. Exception: " + ioe);
      }
    }

  private void writeResults(String fileName)
    {
	try
	  {
	  BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	  StringBuffer sb = new StringBuffer();

	  sb.append("<?xml version = \"1.0\"?>\n");
	  sb.append("<!-- XPB Processing Results -->\n");
	  sb.append("<xpbresults xmlns=\"http://www.xperf.org/xpb/results\">\n");

	  sb.append("  <MeasurementTime millis=\"" + System.currentTimeMillis() + "\"" +
	            " timestamp=\"" + new Date() + "\"/>\n");

      sb.append("  <ProcessingResults>\n");
	  bw.write(sb.toString(), 0, sb.length());
	  bw.flush();

	  for (int i = 0; i < argPMethods.size(); i++)
	    {
	    PMethod pm = (PMethod)argPMethods.elementAt(i);
	    String pData = pm.resultsAsXML(ardr.uris);
	    bw.write(pData, 0, pData.length());
	    bw.flush();
        }

      sb = new StringBuffer();
      sb.append("  </ProcessingResults>\n");
      sb.append("</xpbresults>\n");
      bw.write(sb.toString(), 0, sb.length());
      bw.close();
      }
    catch (IOException ioe)
      {
	  System.err.println("Encountered error in writing reults. Exception: " + ioe);
      }
    }

  private void getArgs(String[] args)
    {
	ardr = new ArgsReader(pActivityMap);
	ardr.readArgs();

    parseCLArgs(args, ardr);	  // Might override some of the flags and parameters.

	if (ardr.verboseFlag)
	  ardr.displayArgs();

	argPMethods = ardr.argPMethods;
    }

  public static void main(String [] args)
    {
	XPBDriver driver = new XPBDriver();

	driver.configure();
	driver.getArgs(args);

	boolean verboseFlag = driver.ardr.verboseFlag;

	if (verboseFlag)
	  System.out.println("Creating Processors ...");
	driver.createProcessors();
	if (verboseFlag)
	  System.out.println("... Created Processors.");

	// Perform the Processing Activities
	if (verboseFlag)
	  System.out.println("Processing Started ...");
	driver.process();
	if (verboseFlag)
	  System.out.println("... Processing Completed.");

	if (verboseFlag)
	  System.out.println("Writing Performance Measurements to \"" + PDATA_FILE + "\" ...");
	driver.writePerfData(PDATA_FILE);

	if (verboseFlag)
	  System.out.println("Writing Processing Results to \"" + RESULTS_FILE + "\" ...");
	driver.writeResults(RESULTS_FILE);

	driver.display();
    System.exit(0);
    }
  }
